
// Region.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2005

/* A region is a piece of text copied from a longer text
   sequence (or region), whose extent is determined either by 
   start and end landmarks or by a pair of matching tags.

   The start/end landmarks are strings (we don't use REs
   in this code, but it would be a useful extension).

   Examples of matching tag pairs are <table></table> and
   <ul></ul>. The user supplies "table" or "ul". The matching end 
   tag will be at the same nesting level as the start tag. For
   example, this prevents <ul> matching with a </ul> tag nested 
   inside it.

   A new region is generated by applying match3() or tagMatch4() methods
   to a larger region or to a sequence.

   A sequence can be read from a Web page (via a URL), 
   from a local text file, or be an ordinary string. If the sequence comes 
   from a Web page or local file, then it's read lazily, a line at a time, 
   as required by the match() method that generates the region.

   This laziness means that the region's text may not be 
   fully stored in the Region object until the match() method needs
   it. The text already used is stored in a StringBuffer object, to allow
   text to be appended to it efficiently, as it's read from the
   sequence.
*/

import java.io.*;


public class Region
{
  private StringBuffer textBuff;   // holds the region's text

  private Sequence seq = null;     // the Sequence object
  private boolean usingSeq;

  private int positionInSeq;   // where this region starts in the sequence
  
  private int len;          /* the length of this region, which may not be
                               known yet if all the sequence has not been read in */
  
  private int lineCount;   // the number of lines read from the sequence


  public Region(String s)
  // create a region using the sequence represented by s
  {
    seq = new Sequence(s);
    usingSeq = true;
    textBuff = new StringBuffer();  // no text read in from the sequence yet
    positionInSeq = 0;
    len = -1;           // since the sequence hasn't been read in
    lineCount = 0;      // no lines read from sequence yet
  } // end of Region()


  public Region(String s, int posnSeq)
  /* Create a region using the string s. posnSeq is the position
     of this string inside a parent sequence. Use 0 if there isn't
     a parent. */
  {
    usingSeq = false;      // not using a Sequence object
    textBuff = new StringBuffer(s);   // all the region text comes from s
    // System.out.println("region 2 textBuff: " + s);
    positionInSeq = posnSeq; 
    len = s.length();     
  } // end of Region()


  public Region(Sequence seq, String s, int posn, int lc)
  // create a region using an existing sequence and string s
  {
    this.seq = seq;
    usingSeq = true;
    textBuff = new StringBuffer(s);
    // System.out.println("region 3 textBuff: " + s);
    positionInSeq = posn;
    len = -1;  // since the sequence hasn't all been read in
    lineCount = lc;
  } // end of Region()


  // -------------------- match on mth landmark ---------------


  public TripleRegion match3(int m, String startLandmark, int n, String endLandmark)
  /* match3() creates a new matching region based on the current region. The
     new region is delimited by the mth start landmark and the nth end landmark 
     after that. 

     In fact, match3() returns three regions, stored in a TripleRegion 
     object holding the 'left', 'matching', and 'rest' regions generated by the match.

     The left region is the region just before the matching one. The rest region
     is the region from just after the mth start landmark to the end of the 
     enclosing region. left+rest == the original region/sequence

     null is returned if no matching region is found.
  */
  {
    // skip over start landmarks until we reach the mth one
    int startPosn = 0;
    int startCount = m;
    while (startCount > 0) {
      startPosn = scanForLandmark(startPosn, startLandmark, false);   // find start of landmark
      if (startPosn == -1)  // landmark not found
        return null;
      startCount--;
      startPosn += startLandmark.length();   // move past the landmark
    }

    // skip over end landmarks until we reach the nth one
    int endPosn = startPosn;
    int endCount = n;
    while (endCount > 0) {
      endPosn = scanForLandmark(endPosn, endLandmark, false);     // find start of landmark
      if (endPosn == -1) // not found
        return null;
      endCount--;
      if (endCount > 0)   // move past the landmark (unless it's the nth one)
        endPosn += endLandmark.length();
    }

    // build triple holding the left, matching, and rest regions
    return buildTriple(startPosn, endPosn);
  }  // end of match3()


  private TripleRegion buildTriple(int mStart, int mEnd)
  /* The matching region extends from positions mStart to mEnd-1
     The left region is 0--mStart-1.
     The rest region is mStart--<end>

     The rest region starts at mStart, so that future searches can
     start from just after the place where the matching region found
     its start landmark.
  */
  {
    // build matching region
    String matchStr = textBuff.substring(mStart, mEnd);
    Region matchingRegion = new Region(matchStr, (positionInSeq + mStart));
      /* positionInSeq fixes the region's position to be relative to the 
         initial sequence. */

    // build left region
    String leftStr = textBuff.substring(0, mStart);
    Region leftRegion = null;
    if (!leftStr.equals(""))
      leftRegion = new Region(leftStr, positionInSeq);

    // build rest region, which may be a sequence or a string
    String restStr = textBuff.substring(mStart);  
    // System.out.println("buildTriple restStr: " + restStr +
    //                  "; " + usingSeq + ";" + lineCount);
    int posn = positionInSeq + mStart;
    Region restRegion = null;
    if (usingSeq)   // the rest region uses a sequence
      restRegion = new Region(seq, restStr, posn, lineCount);
    else if (!restStr.equals(""))   // the rest region is only a string
      restRegion = new Region(restStr, posn);

    // return the triple
    return new TripleRegion(leftRegion, matchingRegion, restRegion);
  }  // end of buildTriple()



  private int scanForLandmark(int startPosn, String landmark, boolean isTag)
  /* Use indexOf() to search through the textBuff StringBuffer looking
     for landmark. The search begins at startPosn, and continues until
     there is no more text.

     The complication is that when the search reaches the end of textBuff,
     there may be more text in the Sequence. So an attempt is made to read 
     in a line of it, and then resume the search. Only if the Sequence 
     object has no more text, is the search abandoned.

     If the search is unsucessful, then -1 is returned. 
  */ 
  {
    int scanPosn = -1;
    boolean tagEnd = false;
    int backupPosn;
    boolean isScanning = true;
    String line;
   
    while (isScanning) {
      scanPosn = textBuff.indexOf(landmark, startPosn);   // scan StringBuffer
      if (isTag)  // the landmark is a tag
        tagEnd = isTagEnd(landmark, scanPosn);
      // System.out.println("scanForLandmark 1: " + scanPosn + 
      //                   "; lineCount: " + lineCount);

      if (scanPosn == -1) {     // landmark not found
        if (usingSeq) {         // but we're using a Sequence object
          line = seq.getLine(lineCount);   // so try reading in the next line
          // System.out.println("scanForLandmark 2: " + line);
          if (line != null) {   // got an extra line
            lineCount++;
            backupPosn = textBuff.length() - landmark.length();   // back up the scan a bit
            if (backupPosn > startPosn)    // back up as little as possible
              startPosn = backupPosn;      // set start position
            // textBuff.append("\n" + line);  // add the new line to the StringBuffer
            textBuff.append(line + "\n");  // add the new line to the StringBuffer
          } 
          else {   // no extra line available, so give up
            usingSeq = false;
            return -1;
          }
        }
        else   // no sequence available
          return -1;
      }
      else if (isTag && !tagEnd)         // match was not the search tag
        startPosn = scanPosn + landmark.length();      // keep searching
      else // landmark found
        isScanning = false;   // stop the scanning
    }
    return scanPosn;
  }  // end of scanForLandmark()


  private boolean isTagEnd(String tag, int scanPosn)
  /* Check if there is a " " or ">" in textBuff at scanPosn+tag's length
     which indicates that the tag ends at that position.
  */
  {
    // System.out.println("isTagEnd 1");
    if (scanPosn == -1)
      return false;

    int endPosn = scanPosn + tag.length();
    if (textBuff.length() >= endPosn) {
      if ((textBuff.charAt(endPosn) == ' ') ||
          (textBuff.charAt(endPosn) == '>'))
        return true;
    }
    return false;
  }  // end of isTagEnd()


  // -------------- other versions of match3() -----------------

  public TripleRegion match3(String startLandmark, String endLandmark)
  // match on first landmark, returning a triple
  {  return match3(1, startLandmark, 1, endLandmark);  }


  public Region match(int m, String startLandmark, int n, String endLandmark)
  /* Return only the matching region delimited by the mth start landmark 
     and the nth end landmark. */
  { TripleRegion r3 = match3(m, startLandmark, n, endLandmark);
    return ((r3 == null) ? null : r3.getMatch());
  }


  public Region match(String startLandmark, String endLandmark)
  // match on first landmark, returning only the matching region
  {  return match(1, startLandmark, 1, endLandmark);  }



  // -------------------- match on mth tag pair ---------------


  public QuatRegion tagMatch4(int m, String tag)
  /* Return a region delimited by the mth tag. 

     Examples of start and end tag pairs are <table></table> and
     <ul></ul>. The user supplies "table" or "ul". The matching end 
     tag will be at the same nesting level as the start tag.

     The start tag is "<" + tag; the end tag is "</" + tag. Note that
     we don't add ">", since this would prevent tag attributes from
     being allowed in the search tag.
  */
  {
    // skip over start tag until we reach the mth one
    int startPosn = 0;
    int startCount = m;
    while (startCount > 0) {
      startPosn = scanForLandmark(startPosn, "<"+tag, true);  // find start tag
      if (startPosn == -1)  // not found
        return null;
      startCount--;
      startPosn += (tag.length()+1);   // move past the landmark; 1 is for the "<";
    }

    int attrStart = startPosn;  // the attributes may be just after the start tag
    int attrEnd = scanForLandmark(attrStart, ">", false);  // find end of start tag
    startPosn = attrEnd+1;  // skip past the "">"

    /* Start searching for the end tag, but it must be at the same level
       at the start tag. Ensure this by increnting and decrementing count
       until the end tag occurs when count == 0. */
    int regionPosn = startPosn; 
    int count = 1;   // 1 is to count the start tag that we just found
    int endPosn = -1;

    while (count > 0) {
      endPosn = scanForLandmark(regionPosn, "</"+tag, true);  // find end tag
      if (endPosn == -1) // end tag not found
        return null;
      else {
        count--;   // found an end tag, so decrement count
        count += countTag(regionPosn, endPosn, "<"+tag);  
           // count any start tags found in the region before the end tag
        regionPosn = endPosn + tag.length()+2; // move past the end tag (2 is for "</")
      }
    }

    // build triple holding the matching region
    return buildQuat(startPosn, endPosn, attrStart, attrEnd);
  }  // end of tagMatch4()



  private int countTag(int startPosn, int endPosn, String tag)
  /* Count the number of occurrences of tag between startPosn and
     endPosn. 

     As with scanForLandmark() a complication is that even if the 
     StringBuffer is fully scanned by countTag(), there may be more
     lines to check in Sequence, so these should be read in, a
     line at a time.
  */
  {
    int scanPosn, backupPosn;
    boolean isScanning = true;
    boolean tagEnd;
    String line;
    int matchCount = 0;
   
    while (isScanning) {
      scanPosn = textBuff.indexOf(tag, startPosn);  // scan for tag
      tagEnd = isTagEnd(tag, scanPosn);

      if (scanPosn == -1) {  // tag not found
        if (usingSeq) {      // but we're using a sequence
          line = seq.getLine(lineCount);   // so read in a new line
          if (line != null) {   // got an extra line
            lineCount++;
            backupPosn = textBuff.length() - tag.length();   // backup the scan
            if (backupPosn > startPosn)    // backup as little as possible
              startPosn = backupPosn;
            textBuff.append("\n" + line);  // add the new line to the StringBuffer
          } 
          else {   // no extra line available, so return current count
            usingSeq = false;
            return matchCount;
          }
        }
        else    // no sequence available
          return matchCount;
      }
      else if (!tagEnd)          // match was not the search tag
        startPosn = scanPosn + tag.length();      // keep searching
      else {  // tag found
        if (scanPosn > endPosn)  // but match was found after end posn
          return matchCount;     // so don't include it in the returned count
        else {   // found match before end posn, so keep counting
          startPosn = scanPosn + tag.length();
          matchCount++;
        }
      }
    }
    return matchCount;
  }  // end of countTag()


  private QuatRegion buildQuat(int mStart, int mEnd, int attrStart, int attrEnd)
  /* The matching region extends from positions mStart to mEnd-1
     The left region is 0--mStart-1.
     The rest region is mStart--<end>

     The attribute region is attrStart -- attrEnd

     The rest region starts at mStart, so that future searches can
     start from just after the place where the matching region found
     its start landmark.
  */
  {
    TripleRegion tr = buildTriple(mStart, mEnd);

    // pull it apart to get the three regions for the QuatRegion
    Region leftRegion = tr.getLeft();
    Region matchingRegion = tr.getMatch();
    Region restRegion = tr.getRest();

    Region attributeRegion = null;
    if (attrStart < attrEnd) {   // there's some attributes (or white space)
      // System.out.println("Attribute range: " + attrStart + "--" + attrEnd);

      // build attribute region
      String attrStr = textBuff.substring(attrStart, attrEnd);
      String as = attrStr.trim();   // remove spaces
      if (!as.equals(""))
        attributeRegion = new Region(as, (positionInSeq + attrStart));
        /* positionInSeq fixes the region's position to be relative to the 
           initial sequence. */
      // System.out.println("Attribute region: \"" + attributeRegion + "\"");
    }

    QuatRegion qr = new QuatRegion(leftRegion, matchingRegion,
                                        attributeRegion, restRegion);
    return qr;
  }  // end of buildQuat()

  // ----------------- other versions of tagMatch4() -------------------

  public QuatRegion tagMatch4(String tag)
  // match on first tag pair, returning a triple
  {  return tagMatch4(1, tag);  }


  public Region tagMatch(int m, String tag)
  /* Return a region delimited by the mth tag. Return only
     the matching region. */
  { QuatRegion qr = tagMatch4(m, tag);
    return ((qr == null) ? null: qr.getMatch());
  }

  public Region tagMatch(String tag)
  // match on first tag pair, returning only the matching region
  {  return tagMatch(1, tag);  }


  // -------------- other public methods ----------------

  public String toString()
  /* Return this region as a string. If the region is not fully
     loaded from the Sequence, then the text must be read in from 
     it first.
  */
  {
    if (usingSeq) {   // part of the text is in a sequence
      String line;
      while ((line = seq.getLine(lineCount)) != null) {
        lineCount++;
        textBuff.append("\n" + line);  // add line to StringBuffer
      }
      usingSeq = false;   // since region is now completelt loaded
    }
    return textBuff.toString();
  }  // end of toString()


  public boolean contains(String s)
  // is s somewhere in this region?
  {  
    // String regionText = toString();
    // return (regionText.indexOf(s) != -1);
    return (scanForLandmark(0, s, false) != -1);  
  }



  public int getPosition()
  // where does this region start in the top-level sequence?
  {  return positionInSeq;  }


  public int getLength()
  // what is the length of this region?
  {  
    if (!usingSeq && (len == -1))
      len = textBuff.length();  // update len, since region is completely loaded
    return len;    // may be -1 if still using Sequence
  } // end of getLength()

}  // end of Region class