
// RegionIterator.java
// Andrew Davison, June 2005, ad@fivedots.coe.psu.ac.th

/* RegionIterator implements the Iterator interface to allow
   the easy iteration through multiple regions extracted
   from a parent region.

   The extracted regions are obtained by applying match() ops.

   Two types of match() operations can be applied: one using
   landmarks, the other using tags. The choice is determined
   when the RegionIterator is instantiated.
*/

import java.util.*;
import java.io.*;


public class RegionIterator implements Iterator 
{
  private Region parentRegion;   
    /* the parent region, which gets gradually smaller as matching
       regions are extracted from it. */ 

  private String startLandmark, endLandmark, tag;
  private boolean usingLandmarks;   // whether using landmarks or tags

  private boolean needMatch = true;
  private Region matchingRegion = null;  // the current matching region


  public RegionIterator(Region r, String sl, String el)
  // create an iterator which will use landmarks to generate regions
  { parentRegion = r;
    startLandmark = sl;
    endLandmark = el;
    usingLandmarks = true;
  } // end of RegionIterator()


  public RegionIterator(Region r, String tag)
  // create an iterator which will use tags to generate regions
  { parentRegion = r;
    this.tag = tag;
    usingLandmarks = false;
  }  // end of RegionIterator()



  public boolean hasNext() 
  /* Return true if another matching region can be
     extracted from the parent. This is 'tested' by doing the extraction
     with findMatch(), but only if needMatch == true.

     findMatch() stores the new match in matchingRegion, and the
     remaining part of the parent is stored in parentRegion.
  */
  { if (needMatch) {
      findMatch();
      needMatch = false;   /* don't need a match anymore, since we've
                              just generated one inside findMatch(). */
    }
    return (matchingRegion != null);
  }  // end of hasNext()



  public Object next() 
  /* Return the current matching region. If needMatch is true, then one
     needs to be generated by calling findMatch(), otherwise we can use
     one calculated earlier (probably when hasNext() was called).
  */
  { 
    // System.out.println("next");
    if (needMatch)
      findMatch();
    needMatch = true;  /* Since we're about to serve up the current match
                          so will need another match for next time. */
    return matchingRegion;
  } // end of next()



  private void findMatch()
  /*
     findMatch() extracts a new matching region from the parent region
     using landmarks or tags. The new match is stored in matchingRegion, 
     and the remaining part of the parent is stored back into parentRegion.
  */
  {
    // System.out.println("findMatch 1: " + parentRegion);
    if (parentRegion == null)
      return;    // give up, since nothing to search

    if (usingLandmarks) {   // match with landmarks
      TripleRegion tr = parentRegion.match3(startLandmark, endLandmark);
      if (tr != null) {   // match found; update globals
        matchingRegion = tr.getMatch();
        parentRegion = tr.getRest();
        // System.out.println("findMatch 2 parentRegion: \"" + parentRegion +"\"");
      }
      else {  // no match found
        // System.out.println("findMatch 3, null");
        matchingRegion = null;    // assuming no match found
        parentRegion = null;
      }
    }   
    else {  // match with tags
      QuatRegion qr = parentRegion.tagMatch4(tag); 
      if (qr != null) {   // match found; update globals
        matchingRegion = qr.getMatch();
        parentRegion = qr.getRest();
        // System.out.println("findMatch 4 parentRegion: \"" + parentRegion +"\"");
      }
      else {  // no match found
        // System.out.println("findMatch 5, null");
        matchingRegion = null;    // assuming no match found
        parentRegion = null;
      }
    }
  }  // end of findMatch()
  

  public void remove() 
  // removal not supported
  {  throw new UnsupportedOperationException(
          "RegionIterator does not support remove()");
  }

} // end of RegionIterator