
// Sequence.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2005

/*  A text sequence can come from three sources: a Web page
    specified with a URL, a local file, or from a string.

    In the first two cases, the text is read in incrementally, a
    line at a time, triggered by calls to getLine().

    If the sequence is just a string, then the first call to getLine()
    returns it all, then returns null after that.

    A Sequence object may be used as the source of text by several
    regions, so it keeps a store (cache) of the lines it's read in.
*/

import java.io.*;
import java.net.*;
import java.util.*;


public class Sequence
{
  // change this proxy information as required
  private static final String PROXY_HOST = "cache.psu.ac.th";
  private static final String PROXY_PORT = "8080";


  private BufferedReader br = null;    // input stream from Web page / file
  private String seqString = null;     // used if the sequence is just a string

  private ArrayList lines;   // list of lines already read in
  private int numLinesRead;  // number of lines already read in
  private boolean allRead = false;
        // have all the lines been read in?


  public Sequence(String s)
  // create a sequence
  {
    boolean isOpen = false;

    if (s.startsWith("http://"))   // may be a URL
      isOpen = openURL(s);

    if (!isOpen && (s.indexOf('.') != -1))   // may be a file
      isOpen = openFile(s);

    if (isOpen) {   // a sequence from a Web page / file needs a lines cache
      lines = new ArrayList();
      numLinesRead = 0;
    }
    else {  // the input string, s, becomes the sequence
      System.out.println("Sequence is : \"" + s + "\"");
      seqString = s;
    }
  }  // end of Sequence()


  private boolean openURL(String urlStr)
  // read text from a URL
  {
    boolean isOpen = false;

    // set up proxy settings; edit as necessary
/*
    Properties sysProps = System.getProperties();
    sysProps.put("proxySet", "true");
    sysProps.put("proxyHost", PROXY_HOST);   // edit these constants to suit you
    sysProps.put("proxyPort", PROXY_PORT);
    System.setProperties(sysProps);
*/
    try {  
      URL u = new URL(urlStr);
      System.out.println("Accessing URL: " + urlStr);
      br = new BufferedReader( new InputStreamReader( u.openStream()));
      isOpen = true;
    } 
    catch(MalformedURLException e) 
    {  System.out.println("Malformed URL: " + urlStr); }
    catch (IOException e) 
    { System.out.println("Cannot read URL: " + urlStr); }

    return isOpen;
  }  // end of openURL()



  private boolean openFile(String fnm)
  // read text from a local file
  {
    boolean isOpen = false;
    try {  
      br = new BufferedReader( new FileReader(fnm));
      System.out.println("Opened file: " + fnm);
      isOpen = true;
    } 
    catch(FileNotFoundException e) 
    {  System.out.println("Could not find file: " + fnm); }
    catch (IOException e) 
    { System.out.println("Cannot read file: " + fnm); }

    return isOpen;
  }  // end of openFile()


  public String getLine(int lineNum)
  /* Return the text for line number lineNum. For a Web page / file
     this may mean downloading another line, or using a cached one.
  */
  {
    if (allRead && (lineNum >= numLinesRead))
      return null;
    
    if (seqString != null) {  // if sequence is a string
      allRead = true;
      numLinesRead = 1;
      return seqString;
    }

    // if we get here then the sequence must be a stream from a file / web page
    if (lineNum < numLinesRead)       // if line already read in
      return (String) lines.get(lineNum);    // then return it

    // if we get here then the line hasn't yet been read in from the stream

    /* A region reads in a line at a time, so a getLine() request will either
       be for a previously cached line, or for the _next_ line in the stream.
       I've used this to simplify the readLine() code below. However, I check
       the requested lineNum value just in case. */

    if (lineNum != numLinesRead)
      System.out.println("**Line Number Error**");   // should not happen

    String line = null;
    try {  
      line = br.readLine();  // try to read in another line
      if (line == null) {    // there was no line to be read
        br.close();
        allRead = true;
      }
      else { // we did read in a new line 
        lines.add(line);   // cache it for later use
        numLinesRead++;
      }
    }
    catch (IOException e) {}
    return line;
  } // end of getLine()


}  // end of Sequence class